from typing import Optional, Tuple

import numpy as np

from oscilloscope.oscilloscope_common import MeasureSettings, compare_ivc
import product.mode as md


kernel = [0] * 21  # [0.05, 0.15, 0.6, 0.15, 0.05]
for i in range(0, 21):
    i2 = (i - 10) * 2.0  # 0.2
    kernel[i] = np.cos(3.14159265 * 0.25 * i2 / 11)
# kernel[10] = 1
summ = 0
for i in range(0, 21):
    summ += kernel[i]
for i in range(0, 21):
    kernel[i] /= summ


def is_in_cluster(ivc, cluster, trh):
    for ivc_cl in cluster:
        s = compare_ivc(ivc, ivc_cl)
        if s < trh:
            return True
    return False


class Pin:
    def __init__(
            self, *positional_args, rc=None, xy=None, ivc=None,
            reference_ivc=None, frequency=None, voltage=None, current=None):
        if len(positional_args) > 0:
            raise NotImplementedError("Positional arguments not supported")

        if rc is not None:
            self.__r = float(rc[0])  # y
            self.__c = float(rc[1])  # x
        elif xy is not None:
            self.__r = float(xy[1])  # y
            self.__c = float(xy[0])  # x
        else:
            self.__r = None
            self.__c = None

        self.__is_out_of_bounds = False

        self.__is_dynamic = False
        self.__cluster_id = -1

        self.__ivc = ivc
        self.__reference_ivc = reference_ivc
        self.__score = None

        self.__frequency = frequency
        self.__voltage = voltage
        self.__current = current

    def __repr__(self):
        return "r = {0}, c = {1}, ivc = {2}".format(self.__r, self.__c, self.__ivc)

    def __setitem__(self, index: int, item: Optional[float]) -> None:
        if index == 0:
            self.__r = float(item)
        elif index == 1:
            self.__c = float(item)
        else:
            raise IndexError("Index out of range")

    def __getitem__(self, index: int) -> Optional[float]:
        if index == 0:
            return self.__r
        elif index == 1:
            return self.__c
        else:
            raise IndexError("Index out of range")

    def to_json(self) -> dict:
        return {
            "x": self.x,
            "y": self.y,
            "is_dynamic": bool(self.is_dynamic),
            "cluster_id": int(self.__cluster_id),
            "ivc": _ivc_to_json(self.ivc),
            "reference_ivc": _ivc_to_json(self.reference_ivc),
            "score": float(self.score) if self.score is not None else None

        }

    @classmethod
    def from_json(cls, d: dict) -> "Pin":
        pin = Pin(
            xy=(float(d["x"]), float(d["y"])),
            ivc=_ivc_from_json(d.get("ivc")),
            reference_ivc=_ivc_from_json(d.get("reference_ivc"))
        )
        pin.__cluster_id = int(d["cluster_id"])
        pin.__score = float(d["score"]) if d["score"] is not None else None  # TODO: Maybe, clean?
        pin.__frequency = md.FrequencyValues[pin.ivc[0].probe_signal_frequency]
        pin.__voltage = md.VoltageValues[pin.ivc[0].max_voltage]
        mf = md.MeasureFlags(pin.ivc[0].flags)

        if mf.current_sense_mode_2_m5_a:
            pin.__current = md.Current.CURRENT_2_5MA
        elif mf.current_sense_mode_25_ma:
            pin.__current = md.Current.CURRENT_25MA
        else:
            pin.__current = md.Current.CURRENT_250UA
        return pin

    @property
    def r(self) -> Optional[float]:
        return self.__r

    @r.setter
    def r(self, value: float):
        self.__r = float(value)

    @r.deleter
    def r(self) -> None:
        self.__r = None

    @property
    def c(self) -> Optional[float]:
        return self.__c

    @c.setter
    def c(self, value: Optional[float]):
        self.__c = float(value)

    @c.deleter
    def c(self) -> None:
        self.__c = None

    x = c
    y = r

    @property
    def rc(self) -> Tuple[Optional[float], Optional[float]]:
        return self.__r, self.__c

    @rc.setter
    def rc(self, value: Optional[Tuple[Optional[float], Optional[float]]]) -> None:
        if value is None:
            self.__r, self.__c = None, None
        else:
            self.r, self.c = value

    @rc.deleter
    def rc(self) -> None:
        self.__r, self.__c = None, None

    @property
    def xy(self) -> Tuple[Optional[float], Optional[float]]:
        return self.__c, self.__r

    @xy.setter
    def xy(self, value: Optional[Tuple[Optional[float], Optional[float]]]) -> None:
        if value is None:
            self.__c, self.__r = None, None
        else:
            self.x, self.y = value

    @xy.deleter
    def xy(self) -> None:
        self.__c, self.__r = None, None

    @property
    def is_out_of_bounds(self):
        return self.__is_out_of_bounds

    @is_out_of_bounds.setter
    def is_out_of_bounds(self, value):
        self.__is_out_of_bounds = bool(value)

    @property
    def ivc(self):
        return self.__ivc

    @ivc.setter
    def ivc(self, value):
        self.__score = None
        if value is None or value[0] is None or value[1] is None:
            self.__ivc = None
        else:
            self.__ivc = value

    @ivc.deleter
    def ivc(self):
        self.__ivc = None
        self.__score = None

    @property
    def reference_ivc(self):
        return self.__reference_ivc

    @reference_ivc.setter
    def reference_ivc(self, value):
        self.__score = None
        if value is None or value[0] is None or value[1] is None:
            self.__reference_ivc = None
        else:
            self.__reference_ivc = value

    @reference_ivc.deleter
    def reference_ivc(self):
        self.__reference_ivc = None
        self.__score = None

    @property
    def is_dynamic(self):
        return self.__is_dynamic

    @is_dynamic.setter
    def is_dynamic(self, value):
        self.__is_dynamic = bool(value)

    @is_dynamic.deleter
    def is_dynamic(self):
        self.__is_dynamic = False

    @property
    def score(self):
        self.calculate_score()
        return self.__score

    @score.deleter
    def score(self):
        self.__score = None

    def calculate_score(self):
        if self.__score is None:
            self.__score = compare_ivc(self.__ivc, self.__reference_ivc)

    def determine_cluster(self, ivc_clusters, trh):
        for i, cluster in enumerate(ivc_clusters):
            if is_in_cluster(self.__ivc, cluster, trh):
                self.__cluster_id = i
                return
        self.__cluster_id = -1

    @property
    def frequency(self):
        return self.__frequency

    @property
    def voltage(self):
        return self.__voltage

    @property
    def current(self):
        return self.__current


def _ivc_to_json(ivc) -> Optional[dict]:
    if ivc is not None and ivc[0] is not None:
        return {
            "measure_settings": ivc[0].to_json(),
            "voltage": tuple(map(float, ivc[1][0])),
            "current": tuple(map(float, ivc[1][1]))
        }


def _ivc_from_json(d: Optional[dict]):
    if d is not None:
        measure_settings = MeasureSettings.from_json(d["measure_settings"])
        voltage = d["voltage"]
        current = d["current"]
        if len(voltage) != len(current) \
                or (len(voltage) != measure_settings.n_points and measure_settings.n_points != 0):
            raise ValueError("Incorrect test points count")

        curve = np.ndarray(shape=(2, len(voltage)))
        curve[0, :] = voltage
        curve[1, :] = current
        return measure_settings, curve
