import numpy as np

from . import ivm
from .oscilloscope_common import MeasureSettings, Oscilloscope


__all__ = ["OscilloscopeURPC"]


_points_per_measure = 25


class OscilloscopeURPC(Oscilloscope):
    def __init__(self, device_name):
        super().__init__()

        self._device = ivm.IvmDeviceHandle(device_name.encode(), defer_open=True)

    @property
    def is_opened(self):
        return self._device._handle is not None

    def open_device(self):
        self.close_device()
        self._device.open_handle()
        self._device.get_status()
        self.get_settings()

    def close_device(self):
        if self.is_opened:
            self._device.close_handle()

    def measure(self):
        self._device.start_measurement()
        while not bool(self._device.measurement_ready().ready_status):
            # sleep(0.002)
            pass

        curve = np.zeros((2, self.s.n_points), dtype=np.float32)
        for frame in range((self.s.n_points - 1) // _points_per_measure + 1):
            measurement = self._device.get_measurement(frame)
            begin = _points_per_measure * frame
            end = _points_per_measure * (frame + 1)
            if end > self.s.n_points:
                end = self.s.n_points
            curve[0, begin: end] = measurement.voltage[: end - begin]
            curve[1, begin: end] = measurement.current[: end - begin]

        curve[1] /= 1000.0  # to A
        curve[:, 0] = curve[:, 1]  # tmp fix for #11468#note-30
        return self.s, curve

    def fast_calibrate(self):
        pass

    def full_calibrate(self):
        self._device.calibrate()

    def set_settings(self, **kwargs):
        super().set_settings(**kwargs)

        mp = self._device.get_measure_settings()

        mp.desc_frequency = self.s.desc_frequency
        mp.probe_signal_frequency = self.s.probe_signal_frequency
        mp.n_points = self.s.n_points
        mp.n_charge_points = self.s.n_charge_points
        mp.max_voltage = self.s.max_voltage
        mp.max_current = self.s.max_current
        mp.measure_flags = self.s.flags
        mp.reserved[:] = self.s.reserved

        self._device.set_measure_settings(mp)

    def get_settings(self):
        mp = self._device.get_measure_settings()
        self.s = MeasureSettings(
            desc_frequency=mp.desc_frequency,
            probe_signal_frequency=mp.probe_signal_frequency,
            n_points=mp.number_points,
            n_charge_points=mp.number_charge_points,
            max_voltage=mp.max_voltage,
            max_current=mp.max_current,
            flags=mp.measure_flags,
            reserved=np.array(mp.reserved))
        return self.s

    def light_control(self, enable):
        if enable:
            self._device.head_light(
                ivm.IvmDeviceHandle.HeadLightRequest.State.HEAD_LIGHT_STATE_ON
            )
        else:
            self._device.head_light(0)
